=== modified file 'drizzled/serialize/Makefile.am'
--- drizzled/serialize/Makefile.am	2008-08-19 07:22:19 +0000
+++ drizzled/serialize/Makefile.am	2008-08-19 08:05:28 +0000
@@ -19,29 +19,43 @@
 
 libserialize_la_CXXFLAGS =
 
+TESTS_ENVIRONMENT = sh
+TESTS = check_length.sh
+
+PROTOC = protoc
+
+%.pb.h %.pb.cc : %.proto
+	$(PROTOC) --cpp_out=${top_builddir}/drizzled/serialize $<
+
+bin_PROGRAMS = \
+	       table_reader \
+	       table_writer \
+               binlog_writer \
+               binlog_reader \
+	       schema_reader
+
+check_PROGRAMS = length
+
 noinst_LTLIBRARIES = libserialize.la
 
 libserialize_la_SOURCES = \
 			  table.pb.cc \
-			  schema.pb.cc 
+			  binary_log.pb.cc \
+			  schema.pb.cc
 
-noinst_HEADERS = \
-		 table.pb.h \
+noinst_HEADERS = table.pb.h \
 		 schema.pb.h \
+		 binary_log.pb.h \
+		 binlog_encoding.h \
 		 serialize.h
 
-schema.pb.h schema.pb.cc: schema.proto
-	protoc --cpp_out=${top_builddir}/drizzled/serialize schema.proto
-
-table.pb.h table.pb.cc: table.proto
-	protoc --cpp_out=${top_builddir}/drizzled/serialize table.proto
-	
 DEFS= @DEFS@
 
-bin_PROGRAMS = \
-	       table_reader \
-	       table_writer \
-	       schema_reader
+binlog_reader_SOURCES = binlog_reader.cc binary_log.cc binlog_encoding.h
+binlog_reader_CXXFLAGS = 
+
+binlog_writer_SOURCES = binlog_writer.cc binary_log.cc binlog_encoding.h
+binlog_writer_CXXFLAGS = 
 
 schema_reader_SOURCES =  schema_reader.cc
 
@@ -51,5 +65,12 @@
 table_writer_SOURCES =  table_writer.cc
 table_writer_CXXFLAGS =
 
-EXTRA_DIST = schema.proto table.proto
-BUILT_SOURCES = schema.pb.h schema.pb.cc table.pb.h table.pb.cc
+length_SOURCES = length.cc binlog_encoding.h
+length_CXXFLAGS = -fexceptions
+
+EXTRA_DIST = schema.proto table.proto binary_log.proto
+BUILT_SOURCES = schema.pb.h \
+		table.pb.h \
+		schema.pb.cc \
+		table.pb.cc \
+		binary_log.pb.cc

=== added file 'drizzled/serialize/binary_log.cc'
--- drizzled/serialize/binary_log.cc	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/binary_log.cc	2008-08-19 07:43:02 +0000
@@ -0,0 +1,156 @@
+#include "binary_log.h"
+
+#include <google/protobuf/io/coded_stream.h>
+
+using namespace google::protobuf;
+using namespace google::protobuf::io;
+
+bool
+BinaryLog::Event::write(CodedOutputStream* out) const
+{
+  // We frame each event in a length encoded in a special manner, and
+  // end it with a CRC-32 checksum.
+
+  // Write length and type
+  unsigned char buf[LENGTH_ENCODE_MAX_BYTES + 1];
+  unsigned char *end= length_encode(m_message->ByteSize(), buf);
+  *end++= m_type;
+
+  char cs[4] = { 0 };                           // !!! No checksum yet
+  if (!out->WriteRaw(buf, end - buf) ||         // Length + Type
+      !m_message->SerializeToCodedStream(out) || // Event body
+      !out->WriteRaw(cs, sizeof(cs)))           // Checksum
+    return false;
+
+  return true;
+}
+
+
+bool
+BinaryLog::Event::read(CodedInputStream *in)
+{
+  unsigned char buf[LENGTH_ENCODE_MAX_BYTES + 1];
+
+  // Read length peek byte to figure out length
+  if (!in->ReadRaw(buf, 1))
+    return false;
+
+  // Read in the rest of the length bytes plus the type
+  size_t bytes= length_decode_bytes(*buf);
+  if (!in->ReadRaw(buf + 1, bytes))
+    return false;
+
+  size_t length;
+  (void) length_decode(buf, &length);
+
+  // Fetch type from read buffer
+  m_type= static_cast<EventType>(buf[bytes]);
+
+  // Create the right event based on the type code (is there something
+  // better in the protobuf library?)
+  Message *message= NULL;
+  switch (m_type) {
+  case QUERY:
+    message= new BinaryLog::Query;
+    break;
+
+  case COMMIT:
+    message= new BinaryLog::Commit;
+    break;
+
+  case ROLLBACK:
+    message= new BinaryLog::Rollback;
+    break;
+
+  case START:
+    message= new BinaryLog::Start;
+    break;
+
+  case CHAIN:
+    message= new BinaryLog::Chain;
+    break;
+  }
+
+  if (!message)
+    return false;
+
+  // Read the event body as length bytes. It is necessary to limit the
+  // stream since otherwise ParseFromCodedStream reads all bytes of
+  // the stream.
+  CodedInputStream::Limit limit= in->PushLimit(length);
+  if (!message->ParseFromCodedStream(in))
+    return false;
+  in->PopLimit(limit);
+  delete m_message;
+  m_message= message;
+
+  // Read checksum (none here yet)
+  char checksum[4];
+  if (!in->ReadRaw(checksum, sizeof(checksum)))
+    return false;
+  return true;
+}
+
+template <class EventClass>
+void print_common(std::ostream& out, EventClass* event)
+{
+  out << "# Global Id: (" << event->header().server_id() << "," << event->header().trans_id() << ")\n";
+}
+
+
+void
+BinaryLog::Event::print(std::ostream& out) const
+{
+  using namespace google::protobuf;
+
+  switch (m_type) {
+  case QUERY:
+  {
+    Query *event= static_cast<Query*>(m_message);
+    print_common(out, event);
+    for (RepeatedPtrField<Query::Variable>::const_iterator ii= event->variable().begin() ;
+         ii != event->variable().end() ;
+         ++ii)
+    {
+      out << "set @" << ii->name() << " = '" << ii->value() << "'\n";
+    }
+    out << event->query() << std::endl;
+    break;
+  }
+
+  case COMMIT:
+  {
+    Commit *event= static_cast<Commit*>(m_message);
+    print_common(out, event);
+    // NYI !!!
+    break;
+  }
+
+  case ROLLBACK:
+  {
+    Rollback *event= static_cast<Rollback*>(m_message);
+    print_common(out, event);
+    // NYI !!!
+    break;
+  }
+
+  case START:
+  {
+    Start *event= static_cast<Start*>(m_message);
+    print_common(out, event);
+    // NYI !!!
+    break;
+  }
+
+  case CHAIN:
+  {
+    Chain *event= static_cast<Chain*>(m_message);
+    print_common(out, event);
+    // NYI !!!
+    break;
+  }
+
+  default:
+    break;                                      /* Nothing */
+  }
+}

=== added file 'drizzled/serialize/binary_log.h'
--- drizzled/serialize/binary_log.h	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/binary_log.h	2008-08-19 07:43:02 +0000
@@ -0,0 +1,60 @@
+// -*- Mode: C++ -*-
+
+#ifndef BINARY_LOG_H_INCLUDED
+#define BINARY_LOG_H_INCLUDED
+
+#include "binary_log.pb.h"
+#include "binlog_encoding.h"
+
+#include <google/protobuf/io/zero_copy_stream.h>
+
+#include <iosfwd>
+#include <stdexcept>
+
+namespace BinaryLog {
+  using namespace google::protobuf;
+  using namespace google::protobuf::io;
+
+  /**
+     Wrapper class to wrap a protobuf event in a type tag and a length.
+
+     The type tag is not part of the actual message, but is handled
+     separately since it is needed to decode the events.
+  */
+  class Event {
+  public:
+    enum EventType {
+      UNDEF,
+      START,
+      CHAIN,
+      COMMIT,
+      ROLLBACK,
+      QUERY,
+      COUNT
+    };
+
+    Event(EventType type, Message *message)
+      : m_type(type), m_message(message)
+    {
+    }
+
+    Event()
+      : m_type(UNDEF), m_message(0)
+    {
+    }
+
+    ~Event() {
+      delete m_message;
+    }
+
+    bool write(CodedOutputStream* out) const;
+    void print(std::ostream& out) const;
+    bool read(CodedInputStream* in);
+
+  private:
+    EventType m_type;
+    Message *m_message;
+  };
+}
+
+#endif /* BINARY_LOG_H_INCLUDED */

=== added file 'drizzled/serialize/binary_log.proto'
--- drizzled/serialize/binary_log.proto	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/binary_log.proto	2008-08-19 07:43:02 +0000
@@ -0,0 +1,52 @@
+package BinaryLog;
+
+// Design goals:
+
+// - Keep the name of the individual binary logs out of the files. We
+//   assume that each binary log has a base name (e.g., the name of
+//   the server) and a sequence number, but we do not want to tie the
+//   base name
+
+message Header {
+  required uint32 server_id = 1;
+  required uint32 trans_id = 2;
+  required uint32 seqno = 3;
+}
+
+// Start of a binary log. Note that the name 
+message Start {
+  required Header header = 1;
+  required uint32 server_version = 2;
+  required string server_signature = 3;
+}
+
+// Chain a binary log to the next binary log
+message Chain {
+  required Header header = 1;
+  required uint32 next = 2;            // Sequence number of next file
+}
+
+message Query {
+  // Assignments to variables that are part of the query
+  message Variable {
+    required string name = 1;
+    required string value = 2;
+    // Character set?
+  }
+
+  required Header header = 1;
+  repeated Variable variable = 3;
+  required string query = 2;
+}
+
+message Commit {
+  required Header header = 1;
+}
+
+message Rollback {
+  required Header header = 1;
+}
+
+
+
+

=== added file 'drizzled/serialize/binlog_encoding.h'
--- drizzled/serialize/binlog_encoding.h	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/binlog_encoding.h	2008-08-19 07:43:02 +0000
@@ -0,0 +1,80 @@
+#ifndef BINLOG_ENCODE_H_INCLUDED
+#define BINLOG_ENCODE_H_INCLUDED
+
+#include <cstdlib>
+#include <cassert>
+#include <cstring>
+#include <stdint.h>
+
+#define LENGTH_ENCODE_MAX_BYTES (sizeof(size_t) + 1)
+
+inline unsigned char *
+length_encode(size_t length, unsigned char *buf)
+{
+  unsigned char *ptr= buf;
+  assert(length > 1);
+  if (length < 256)
+    *ptr++= length & 0xFF;
+  else {
+    int_fast8_t log2m1= -1;        // ceil(log2(ptr - buf)) - 1
+    uint_fast8_t pow2= 1;          // pow2(log2m1 + 1)
+    while (length > 0) {
+      // Check the invariants
+      assert(pow2 == (1 << (log2m1 + 1)));
+      assert((ptr - buf) <= (1 << (log2m1 + 1)));
+
+      // Write the least significant byte of the current
+      // length. Prefix increment is used to make space for the first
+      // byte that will hold log2m1.
+      *++ptr= length & 0xFF;
+      length >>= 8;
+
+      // Ensure the invariant holds by correcting it if it doesn't,
+      // that is, the number of bytes written is greater than the
+      // nearest power of two.
+      if (ptr - buf > pow2) {
+        ++log2m1;
+        pow2 <<= 1;
+      }
+    }
+    // Clear the remaining bytes up to the next power of two
+    memset(ptr + 1, 0, pow2 - (ptr - buf));
+    *buf= log2m1;
+    ptr= buf + pow2 + 1;
+  }
+  return ptr;
+}
+
+inline unsigned char *
+length_decode(unsigned char *buf, size_t *plen)
+{
+  if (*buf > 1) {
+    *plen = *buf;
+    return buf + 1;
+  }
+
+  size_t bytes= 1 << (*buf + 1);
+  unsigned char *ptr= buf + 1;
+  size_t length= 0;
+  for (unsigned int i = 0 ; i < bytes ; ++i)
+    length |= *ptr++ << (8 * i);
+  *plen= length;
+  return ptr;
+}
+
+/**
+   Compute how many bytes are use for the length.
+
+   The number of bytes that make up th length can be computed based on
+   the first byte of the length field. By supplying this byte to the
+   function, the number of bytes that is needed for the length is
+   computed.
+
+ */
+inline size_t
+length_decode_bytes(int peek)
+{
+  return (peek < 2) ? (1 << (peek + 1)) + 1 : 1;
+}
+
+#endif /* BINLOG_ENCODE_H_INCLUDED */

=== added file 'drizzled/serialize/binlog_reader.cc'
--- drizzled/serialize/binlog_reader.cc	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/binlog_reader.cc	2008-08-19 07:43:02 +0000
@@ -0,0 +1,76 @@
+#include "binary_log.h"
+
+#include <iostream>
+
+#include <google/protobuf/io/coded_stream.h>
+#include <google/protobuf/io/zero_copy_stream_impl.h>
+
+#include <getopt.h>
+#include <fcntl.h>
+
+#include <sys/stat.h>
+
+using namespace google::protobuf;
+using namespace google::protobuf::io;
+
+void print_usage_and_exit(char *prog) {
+  using std::cerr;
+  const char *name= strrchr(prog, '/');
+
+  if (name)
+    ++name;
+  else
+    name= "binlog_reader";
+  cerr << "Usage: " << name << " <options>\n"
+       << "    --input name   Read queries from file <name> (default: 'log.bin')\n"
+       << std::flush;
+  exit(1);
+}
+
+
+void
+print_event(BinaryLog::Event *event)
+{
+}
+
+
+int
+main(int argc, char *argv[])
+{
+  using std::ios;
+
+  static struct option options[] = {
+    { "input",     1 /* has_arg */, NULL, 0 },
+    { 0, 0, 0, 0 }
+  };
+
+  const char *file_name= "log.bin";
+
+  int ch, option_index;
+  while ((ch= getopt_long(argc, argv, "", options, &option_index)) != -1) {
+    if (ch == '?')
+      print_usage_and_exit(argv[0]);
+
+    switch (option_index) {
+    case 0:                                     // --input
+      file_name= optarg;
+      break;
+
+    }
+  }
+
+  if (optind > argc)
+    print_usage_and_exit(argv[0]);
+
+  int fd = open(file_name, O_RDONLY);
+  ZeroCopyInputStream* raw_input = new FileInputStream(fd);
+  CodedInputStream *coded_input = new CodedInputStream(raw_input);
+
+  BinaryLog::Event event;
+  while (event.read(coded_input))
+    event.print(std::cout);
+
+  delete coded_input;
+  delete raw_input;
+  close(fd);
+}

=== added file 'drizzled/serialize/binlog_writer.cc'
--- drizzled/serialize/binlog_writer.cc	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/binlog_writer.cc	2008-08-19 07:43:02 +0000
@@ -0,0 +1,135 @@
+#include "binlog_encoding.h"
+#include "binary_log.h"
+
+#include "ioutil.h"
+
+#include <google/protobuf/io/zero_copy_stream_impl.h>
+#include <google/protobuf/io/coded_stream.h>
+
+#include <iostream>
+#include <sstream>
+#include <string>
+#include <map>
+
+#include <getopt.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+
+using namespace google::protobuf::io;
+
+typedef std::map<std::string,std::string> Assign;
+
+void print_usage_and_exit(char *prog) {
+  using std::cerr;
+  const char *name= strrchr(prog, '/');
+
+  if (name)
+    ++name;
+  else
+    name= "binlog_writer";
+  cerr << "Usage: " << name << " <options> <query>\n"
+       << "    --output name    Append query to file <name> (default: 'log.bin')\n"
+       << "    --set var=val    Set value of user variable for query\n"
+       << "    --trans-id <id>  Set transaction id to <id>\n"
+       << std::flush;
+  exit(1);
+}
+
+
+void
+write_query(CodedOutputStream* out,
+            unsigned long trans_id,
+            const std::string& query,
+            const Assign& assign)
+{
+  BinaryLog::Query *message = new BinaryLog::Query;
+
+  {
+    BinaryLog::Header *header= message->mutable_header();
+    header->set_seqno(time(NULL));
+    header->set_server_id(1);
+    header->set_trans_id(trans_id);
+  }
+
+  message->set_query(query);
+  for (Assign::const_iterator ii= assign.begin() ;
+       ii != assign.end() ;
+       ++ii )
+  {
+    BinaryLog::Query::Variable *var= message->add_variable();
+    var->set_name(ii->first);
+    var->set_value(ii->second);
+  }
+
+  BinaryLog::Event event(BinaryLog::Event::QUERY, message);
+  event.write(out);
+}
+
+
+int main(int argc, char *argv[])
+{
+  using std::ios;
+  GOOGLE_PROTOBUF_VERIFY_VERSION;
+
+  static struct option options[] = {
+    { "set",       1 /* has_arg */, NULL, 0 },
+    { "trans-id",  1 /* has_arg */, NULL, 0 },
+    { "output",    1 /* has_arg */, NULL, 0 },
+    { 0, 0, 0, 0 }
+  };
+
+  Assign assign;
+  unsigned long trans_id= 0;
+  const char* file_name= "log.bin";
+
+  int ch, option_index;
+  while ((ch= getopt_long(argc, argv, "", options, &option_index)) != -1) {
+    if (ch == '?')
+      print_usage_and_exit(argv[0]);
+
+    switch (option_index) {
+    case 0:                                     // --set
+    {
+      // Split the supplied string at the first '='
+      char *end= optarg + strlen(optarg);
+      char *pos= strchr(optarg, '=');
+      if (!pos)
+        pos= end;
+      const std::string key(optarg, pos);
+      const std::string value(pos == end ? end : pos+1, end);
+      assign[key]= value;
+    }
+
+    case 1:                                     // --trans-id
+      trans_id= strtoul(optarg, NULL, 0);
+      break;
+
+    case 2:                                     // --output
+      file_name= optarg;
+      break;
+    }
+  }
+
+  if (optind >= argc)
+    print_usage_and_exit(argv[0]);
+
+  int outfd= open(file_name, O_WRONLY | O_CREAT | O_APPEND, 0644);
+  if (outfd < 0)
+  {
+    perror("open");
+    exit(1);
+  }
+
+  ZeroCopyOutputStream* raw_output = new FileOutputStream(outfd);
+  CodedOutputStream* coded_output = new CodedOutputStream(raw_output);
+
+  std::stringstream sout;
+  sout << ioutil::join(" ", &argv[optind], &argv[argc]);
+
+  write_query(coded_output, trans_id, sout.str(), assign);
+
+  delete coded_output;
+  delete raw_output;
+  close(outfd);
+  exit(0);
+}

=== added file 'drizzled/serialize/check_length.sh'
--- drizzled/serialize/check_length.sh	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/check_length.sh	2008-08-19 07:43:02 +0000
@@ -0,0 +1,38 @@
+#!/bin/sh
+
+failed=0
+
+check_encode () {
+    bytes=`./length --hex --encode $1`
+    shift
+    expected="$@"
+    if test "$bytes" != "$expected"; then
+        failed=`expr $failed + 1`
+        echo "Got '$bytes' but expected '$expected'"
+    fi
+}
+
+check_invertible () {
+    bytes=`./length --encode $value`
+    result=`./length --decode $bytes`
+    if test $result -ne $value; then
+        failed=`expr $failed + 1`
+        echo "Got $result but expected $value"
+    fi
+}
+
+# check that the encoding is correct
+check_encode 2     0x2
+check_encode 255   0xff
+check_encode 256   0x0 0x0 0x1
+check_encode 65535 0x0 0xff 0xff
+
+# Check that we can decode what we have encoded
+for value in 2 255 256 65535 65536 \
+    `expr 16 "*" 65536` \
+    `expr 16 "*" 65536 + 255`
+do
+    check_invertible $value
+done
+
+exit `test $failed -gt 0`

=== added file 'drizzled/serialize/ioutil.h'
--- drizzled/serialize/ioutil.h	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/ioutil.h	2008-08-19 07:43:02 +0000
@@ -0,0 +1,67 @@
+// -*- Mode: C++ -*-
+
+#include <iosfwd>
+#include <string>
+
+namespace ioutil {
+
+  /**
+     Helper class for join() I/O manipulator.
+
+     This is a join I/O manipulator with arguments for @c join(). These has to be
+  */
+  template <class FwdIter> class joiner {
+    friend std::ostream& operator<<(std::ostream& out, const joiner& j) {
+      j.write(out);
+      return out;
+    }
+
+  public:
+    explicit joiner(const std::string& separator, FwdIter start, FwdIter finish)
+      : m_sep(separator), m_start(start), m_finish(finish)
+    { }
+
+
+  private:
+    std::string m_sep;
+    FwdIter m_start, m_finish;
+
+    void write(std::ostream& out) const {
+      if (m_start == m_finish)
+        return;
+
+      FwdIter fi = m_start;
+      while (true) {
+        out << *fi;
+        if (++fi == m_finish)
+          break;
+        out << m_sep;
+      }
+    }
+  };
+
+
+  /**
+     Join manipulators for writing delimiter-separated strings to an
+     ostream object.
+
+     Use the manipulator as follows:
+     @code
+     std::cout << ioutil::join(",", list.begin(), list.end()) << std::endl;
+     std::cout << ioutil::join(",", list) << std::endl;
+     @endcode
+  */
+  template <class FwdIter>
+  joiner<FwdIter>
+  join(const std::string& delim, FwdIter start, FwdIter finish) {
+    return joiner<FwdIter>(delim, start, finish);
+  }
+
+  /** @overload */
+  template <class Container>
+  joiner<typename Container::const_iterator>
+  join(const std::string& delim, Container seq) {
+    typedef typename Container::const_iterator FwdIter;
+    return joiner<FwdIter>(delim, seq.begin(), seq.end());
+  }
+}

=== added file 'drizzled/serialize/length.cc'
--- drizzled/serialize/length.cc	1970-01-01 00:00:00 +0000
+++ drizzled/serialize/length.cc	2008-08-19 07:43:02 +0000
@@ -0,0 +1,134 @@
+
+#include "binlog_encoding.h"
+
+#include <cstdlib>
+#include <iostream>
+#include <iomanip>
+#include <stdexcept>
+#include <getopt.h>
+
+using std::ios;
+using std::cout;
+using std::cerr;
+using std::flush;
+
+void print_usage_and_exit(char *prog) {
+  const char *name= strrchr(prog, '/');
+  if (name)
+    ++name;
+  else
+    name= "length";
+  cerr << "Usage:\n"
+       << "    " << name << " [ -vvvx ] -e <number> ...\n";
+  cerr << "    " << name << " [ -vvvx ] -d <byte> ...\n";
+  cerr << flush;
+  exit(1);
+}
+
+void encode(int argc, char *argv[], int verbose_level, bool hex_output) {
+  for (int i = 0 ; i < argc ; ++i) {
+    size_t length = strtoul(argv[i], NULL, 0);
+
+    if (length < 2)
+      throw std::invalid_argument("Length has to be > 1");
+
+    unsigned char buf[128];
+    unsigned char *end= length_encode(length, buf);
+    ios::fmtflags saved_flags= cout.flags();
+    if (verbose_level > 0)
+      cout << "Length " << length << ": ";
+    if (hex_output)
+      cout << std::hex << std::setw(2) << std::setfill('0');
+    unsigned char *ptr= buf;
+    while (true) {
+      if (hex_output)
+        cout << "0x";
+      cout << (unsigned int) *ptr;
+      if (++ptr == end)
+        break;
+      cout << " ";
+    }
+    cout << std::endl;
+    cout.setf(saved_flags);
+  }
+}
+
+
+void decode(int argc, char *argv[], int verbose_level, bool hex_output) {
+  unsigned char buf[128];
+  for (int i = 0 ; i < argc ; ++i)
+    buf[i]= strtoul(argv[i], NULL, 0);
+
+  size_t length;
+  (void) length_decode(buf, &length);
+
+  ios::fmtflags saved_flags= cout.flags();
+  if (verbose_level > 0)
+    cout << "Length ";
+  if (hex_output)
+    cout.setf(ios::hex, ios::basefield);
+  cout << length << std::endl;
+  cout.setf(saved_flags);
+}
+
+
+int main(int argc, char *argv[]) {
+  enum { NO_ACTION, ENCODE_ACTION, DECODE_ACTION } action= NO_ACTION;
+
+  static struct option long_options[] = {
+    { "decode",  0 /* has_arg */, NULL, 'd' },
+    { "encode",  0 /* has_arg */, NULL, 'e' },
+    { "verbose", 0 /* has_arg */, NULL, 'v' },
+    { "hex",     0 /* has_arg */, NULL, 'x' },
+    { 0, 0, 0, 0 }
+  };
+
+  int verbose_level= 0;
+  bool hex_output= false;
+  int ch;
+
+  while ((ch= getopt_long(argc, argv, "devx", long_options, NULL)) != -1) {
+    switch (ch) {
+    case 0:
+    case '?':
+      print_usage_and_exit(argv[0]);
+      break;
+
+    case 'd':
+      action= DECODE_ACTION;
+      break;
+
+    case 'e':
+      action= ENCODE_ACTION;
+      break;
+
+    case 'v':
+      ++verbose_level;
+      break;
+
+    case 'x':
+      hex_output= true;
+      break;
+    }
+  }
+
+  try {
+    switch (action) {
+    case ENCODE_ACTION:
+      encode(argc - optind, argv + optind, verbose_level, hex_output);
+      break;
+    case DECODE_ACTION:
+      decode(argc - optind, argv + optind, verbose_level, hex_output);
+      break;
+    default:
+      print_usage_and_exit(argv[0]);
+      break;
+    }
+  }
+  catch (std::invalid_argument& ex) {
+    cerr << ex.what() << "\n";
+    print_usage_and_exit(argv[0]);
+  }
+
+  return EXIT_SUCCESS;
+}

